@@ -299,7 +299,7 @@ class QuantitativeAssessment(object):
         # Assemble metabolite and reaction data in two data frames.
         self._metabolites_df = pd.DataFrame(
             {
-                "id": pmo_problem.F.metabolite_ids,
+                "id": pmo_problem.T.metabolite_ids,
                 "conc": np.exp(pmo_problem.log_c).ravel() * 1000
                 if pmo_problem.log_c is not None
                 else np.nan,
