@@ -5,7 +5,11 @@ problems.
 from __future__ import annotations
 
 import logging
-from typing import Any, Callable, Dict, Optional, Union
+import multiprocessing
+import os
+import pickle
+import tempfile
+from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
 
 import cvxpy as cp
 import numpy as np
@@ -14,10 +18,89 @@ from scipy.stats.distributions import chi2
 
 import cobra
 
+from .constants import default_confidence_level, default_max_drg, default_min_drg
 from .flux_space import FluxSpace
 from .thermodynamic_space import ThermodynamicSpace, ThermodynamicSpaceBasis
 
 logger = logging.getLogger(__name__)
+_pmo_problem: PmoProblem
+
+
+class PmoProblemPool:
+    """Creation of a process pool for solving multiple PMO problems on the same model."""
+
+    def __init__(self, num_processes: Optional[int], *argv):
+        """Initialize a process pool for solving multiple PMO problems.
+
+        Parameters
+        ----------
+        num_processes : Optional[int]
+            Number of workers in the pool.
+        argv
+            Arguments to be passed to the constructor of :code:`PmoProblem`
+        """
+        # Pick the number of processes based on the number of available CPUs.
+        if num_processes is None:
+            num_processes = multiprocessing.cpu_count()
+            if num_processes is None:
+                logger.warning(
+                    "Cannot determine the number of processors available. Assuming 1."
+                )
+                num_processes = 1
+        assert num_processes >= 1, "At least one process is needed."
+
+        # Create a temporary file to copy the initialization arguments to the worker
+        # processes. Helps performance on Windows.
+        with tempfile.NamedTemporaryFile(delete=False) as args_file:
+            pickle.dump(argv, open(args_file.name, "wb"))
+            self._args_filename = args_file.name
+
+        # Create the process pool, initializing a PmoProblem on each worker.
+        self._pool = multiprocessing.Pool(
+            num_processes,
+            initializer=PmoProblemPool._init_process,
+            initargs=[self._args_filename],
+        )
+
+    def map(
+        self, fn: Callable[[PmoProblem, Any], Any], inputs: Iterable[Any]
+    ) -> List[Any]:
+        """Executes a function on each input element. This is the same as the regular
+        :code:`map` function, except that it executes in parallel on the available
+        workers.
+
+        Parameters
+        ----------
+        fn : Callable[[PmoProblem, Any], Any]
+            Function to be applied to each element.
+        inputs : Iterable[Any]
+            An iterable containing the input elements.
+
+        Returns
+        -------
+        List[Any]
+            A list containing the result of applying the function to each input element.
+        """
+        inputs_with_fn = [(fn, i) for i in inputs]
+        return self._pool.map(PmoProblemPool._pool_task, inputs_with_fn)
+
+    def close(self):
+        """Waits for all jobs to be done and closes the pool."""
+        self._pool.close()
+        self._pool.join()
+        if os.path.isfile(self._args_filename):
+            os.remove(self._args_filename)
+
+    @staticmethod
+    def _init_process(args_filename):
+        global _pmo_problem
+        argv = pickle.load(open(args_filename, "rb"))
+        _pmo_problem = PmoProblem(*argv)
+
+    @staticmethod
+    def _pool_task(v: Tuple[Callable[[PmoProblem, Any], Any], Any]):
+        global _pmo_problem
+        return v[0](_pmo_problem, v[1])
 
 
 class PmoProblem(object):
@@ -46,31 +129,25 @@ class PmoProblem(object):
         Maximum magnitude for the reaction energy of each reaction, by default 1000.
     solver : Optional[str], optional
         Name of the solver to use, this can be any of the solvers supported by CVXPY, by
-        default None. 
+        default None.
     solver_options : dict, optional
         Dictionary specifying additional options for the solver.
     """
 
     _MAX_COEFFICIENT_RANGE = 5e8
-    _DEFAULT_SOLVER_OPTIONS = {
-        cp.GUROBI: {
-            'IntFeasTol': 1e-9,
-            'FeasibilityTol': 1e-9
-        }
-    }
+    _DEFAULT_SOLVER_OPTIONS = {cp.GUROBI: {"IntFeasTol": 1e-9, "FeasibilityTol": 1e-9}}
 
     def __init__(
         self,
         network: Union[cobra.Model, FluxSpace],
         thermodynamic_space: ThermodynamicSpace,
         thermodynamic_space_basis: ThermodynamicSpaceBasis = None,
-        objective: Callable[
-            [PmoProblem], cp.problems.objective.Objective] = None,
-        confidence_level: float = 0.95,
-        min_drg: float = 1e-1,
-        max_drg: float = 1000,
+        objective: Callable[[PmoProblem], cp.problems.objective.Objective] = None,
+        confidence_level: float = default_confidence_level,
+        min_drg: float = default_min_drg,
+        max_drg: float = default_max_drg,
         solver: Optional[str] = None,
-        solver_options: dict = {}
+        solver_options: dict = {},
     ):
         if isinstance(network, FluxSpace):
             self._F = network.copy()
@@ -81,13 +158,16 @@ class PmoProblem(object):
             self._B = thermodynamic_space_basis
         else:
             self._B = ThermodynamicSpaceBasis(thermodynamic_space)
-        assert self.B.to_drg is not None, 'The thermodynamic space basis ' \
-            'must have explicit reaction energies for PMO.'
+        assert self.B.to_drg is not None, (
+            "The thermodynamic space basis "
+            "must have explicit reaction energies for PMO."
+        )
 
         # Find mapping from reactions in F to reactions in T (T only covers a subset of
         # the reactions in F).
         self._rxn_idxs_F_to_T = [
-            self.F.reaction_ids.index(id) for id in self.T.reaction_ids]
+            self.F.reaction_ids.index(id) for id in self.T.reaction_ids
+        ]
 
         # Scale the flux space to make the problem numerically nicer.
         self._flux_scale = PmoProblem._scale_flux_space(self.F)
@@ -102,24 +182,31 @@ class PmoProblem(object):
         self.solver_options = solver_options
 
         # Flux, thermodynamic and direction variables.
-        self._vs = cp.Variable((self.F.n_reactions, 1), 'fluxes')
-        self._m = cp.Variable((self.B.dimensionality, 1), 'm')
-        if np.all(np.logical_or(
+        self._vs = cp.Variable((self.F.n_reactions, 1), "fluxes")
+        self._m = cp.Variable((self.B.dimensionality, 1), "m")
+        if np.all(
+            np.logical_or(
                 self.F.lb[self._rxn_idxs_F_to_T] >= 0,
-                self.F.ub[self._rxn_idxs_F_to_T] <= 0)):
+                self.F.ub[self._rxn_idxs_F_to_T] <= 0,
+            )
+        ):
             # If flux bounds already fix all the directions we do not need integer
             # variables.
             self._fixed_directions = True
-            self._d = cp.Variable((len(self.T.reaction_ids), 1), 'directions')
+            self._d = cp.Variable((len(self.T.reaction_ids), 1), "directions")
         else:
             self._fixed_directions = False
-            self._d = cp.Variable((len(self.T.reaction_ids), 1),
-                                  'directions', boolean=True)
+            self._d = cp.Variable(
+                (len(self.T.reaction_ids), 1), "directions", boolean=True
+            )
 
         # If no objective is specified, maximize the probability of the
         # thermodynamic variables.
-        self.objective = objective or (lambda p: cp.Minimize(
-            cp.atoms.quad_form(p.m, np.identity(p.B.dimensionality))))
+        self._objective = objective or (
+            lambda p: cp.Minimize(
+                cp.atoms.quad_form(p.m, np.identity(p.B.dimensionality))
+            )
+        )
 
         self._make_big_M_coefficients()
         self._make_constraints()
@@ -146,16 +233,17 @@ class PmoProblem(object):
         return self._flux_scale
 
     @property
-    def objective(self) -> Optional[Callable[
-            [PmoProblem], cp.problems.objective.Objective]]:
+    def objective(
+        self,
+    ) -> Optional[Callable[[PmoProblem], cp.problems.objective.Objective]]:
         """Gets the function that constructs the PMO objective."""
         return self._objective
 
     @objective.setter
-    def objective(self, value: Optional[Callable[
-            [PmoProblem], cp.problems.objective.Objective]]):
+    def objective(self, value: Callable[[PmoProblem], cp.problems.objective.Objective]):
         """Sets the function that constructs the PMO objective."""
         self._objective = value
+        self._rebuild_cvxpy_problem()
 
     @property
     def vs(self) -> cp.Variable:
@@ -237,17 +325,17 @@ class PmoProblem(object):
     @property
     def flux_lb_constraint(self) -> cp.constraints.constraint.Constraint:
         """Gets the lower bound constraint for fluxes."""
-        return self._flux_lb_constraint,
+        return (self._flux_lb_constraint,)
 
     @property
     def flux_ub_constraint(self) -> cp.constraints.constraint.Constraint:
         """Gets the upper bound constraint for fluxes."""
-        return self._flux_ub_constraint,
+        return (self._flux_ub_constraint,)
 
     @property
     def steady_state_constraint(self) -> cp.constraints.constraint.Constraint:
         """Gets the steady state constraint."""
-        return self._steady_state_constraint,
+        return (self._steady_state_constraint,)
 
     @property
     def confidence_constraint(self) -> cp.constraints.constraint.Constraint:
@@ -256,7 +344,7 @@ class PmoProblem(object):
 
     @property
     def sign_constraints(self) -> cp.constraints.constraint.Constraint:
-        """Gets the reaction direction constraints. The constraints at indices 
+        """Gets the reaction direction constraints. The constraints at indices
         0 and 1 are constraints on the reaction energies, while 2 and 3 are
         constraints on the fluxes."""
         return self._sign_constraints
@@ -286,7 +374,7 @@ class PmoProblem(object):
 
     def rebuild_for_directions(self, directions: np.array) -> PmoProblem:
         """Construct a copy of this PMO problem constrained to the given reaction
-        directions. 
+        directions.
 
         Parameters
         ----------
@@ -299,13 +387,26 @@ class PmoProblem(object):
         PmoProblem
             A copy of this problem, constrained to the given directions.
         """
-        reaction_idxs = [
-            self.F.reaction_ids.index(id) for id in self.T.reaction_ids]
+        reaction_idxs = [self.F.reaction_ids.index(id) for id in self.T.reaction_ids]
 
-        new_lb = self.F.lb[reaction_idxs, ]
-        new_ub = self.F.ub[reaction_idxs, ]
-        new_lb[directions > 0.5, ] = np.maximum(new_lb[directions > 0.5, ], 0)
-        new_ub[directions < 0.5, ] = np.minimum(new_ub[directions < 0.5, ], 0)
+        new_lb = self.F.lb[
+            reaction_idxs,
+        ]
+        new_ub = self.F.ub[
+            reaction_idxs,
+        ]
+        new_lb[directions > 0.5,] = np.maximum(
+            new_lb[
+                directions > 0.5,
+            ],
+            0,
+        )
+        new_ub[directions < 0.5,] = np.minimum(
+            new_ub[
+                directions < 0.5,
+            ],
+            0,
+        )
 
         constrained_F = self.F.copy()
         constrained_F.lb[reaction_idxs] = new_lb
@@ -320,7 +421,8 @@ class PmoProblem(object):
             self.min_drg,
             self.max_drg,
             self.solver,
-            self.solver_options)
+            self.solver_options,
+        )
 
     @staticmethod
     def _scale_flux_space(F: FluxSpace) -> np.array:
@@ -340,12 +442,15 @@ class PmoProblem(object):
 
         row_scale = np.max(np.abs(F.S), axis=1)
         if any(row_scale == 0):
-            zero_row_mets = [F.metabolite_ids[i] for
-                             i in np.where(row_scale == 0)[0].tolist()]
-            raise Exception('One row of the stoichiometric matrix is zero, '
-                            'Please make sure that the model does not contain orphan '
-                            'metabolites. The metabolites corresponding to zero rows are '
-                            f"{','.join(zero_row_mets)}.")
+            zero_row_mets = [
+                F.metabolite_ids[i] for i in np.where(row_scale == 0)[0].tolist()
+            ]
+            raise Exception(
+                "One row of the stoichiometric matrix is zero, "
+                "Please make sure that the model does not contain orphan "
+                "metabolites. The metabolites corresponding to zero rows are "
+                f"{','.join(zero_row_mets)}."
+            )
         F.S = np.diag(reference / row_scale) @ F.S
 
         return col_scale
@@ -358,38 +463,42 @@ class PmoProblem(object):
         ones = np.ones((len(self.T.reaction_ids), 1))
 
         self._big_M_r = self.max_drg * ones
-        self._big_M_f = 1.01 * np.maximum(
-            np.abs(self.F.lb), np.abs(self.F.ub))[self._rxn_idxs_F_to_T, :]
+        self._big_M_f = (
+            1.01
+            * np.maximum(np.abs(self.F.lb), np.abs(self.F.ub))[self._rxn_idxs_F_to_T, :]
+        )
         self._epsilon_r = self.min_drg * ones
         self._epsilon_f = np.maximum(
             self._big_M_f / self._MAX_COEFFICIENT_RANGE,
-            np.max(self._big_M_f) / self._MAX_COEFFICIENT_RANGE
+            np.max(self._big_M_f) / self._MAX_COEFFICIENT_RANGE,
         )
 
         all_bounds = np.abs(np.vstack((self.F.lb, self.F.ub)))
         min_flux_bound = np.min(all_bounds[np.nonzero(all_bounds)])
 
-        if min_flux_bound < np.max(self._big_M_f) \
-                / self._MAX_COEFFICIENT_RANGE * 10:
-            logger.warning('The coefficient range of the flux bounds '
-                           f'({np.max(self._big_M_f) / min_flux_bound}) is close or '
-                           f'larger than {self.MAX_COEFFICIENT_RANGE} '
-                           'and may lead to numerical errors. It is recommended to '
-                           'simplify the model to reduce the range.')
+        if min_flux_bound < np.max(self._big_M_f) / self._MAX_COEFFICIENT_RANGE * 10:
+            logger.warning(
+                "The coefficient range of the flux bounds "
+                f"({np.max(self._big_M_f) / min_flux_bound}) is close or "
+                f"larger than {self.MAX_COEFFICIENT_RANGE} "
+                "and may lead to numerical errors. It is recommended to "
+                "simplify the model to reduce the range."
+            )
 
     def _make_constraints(self):
-        """Create the CVXPY constraints of the problem.
-        """
+        """Create the CVXPY constraints of the problem."""
         # Flux constraints.
         self._flux_lb_constraint = self.vs >= self.F.lb
         self._flux_ub_constraint = self.vs <= self.F.ub
-        self._steady_state_constraint = (
-            self.F.S @ self.vs == np.zeros((self.F.n_metabolites, 1)))
+        self._steady_state_constraint = self.F.S @ self.vs == np.zeros(
+            (self.F.n_metabolites, 1)
+        )
 
         # Thermodynamic space constraints.
         CI_square = chi2.ppf(self._confidence_level, self.B.dimensionality)
-        self._m_constraint = cp.atoms.quad_form(
-            self.m, np.identity(self.B.dimensionality)) <= CI_square
+        self._m_constraint = (
+            cp.atoms.quad_form(self.m, np.identity(self.B.dimensionality)) <= CI_square
+        )
 
         # Mappings between fluxes and constrained fluxes.
         v_to_v_con = np.identity(self.F.n_reactions)[self._rxn_idxs_F_to_T, :]
@@ -399,19 +508,23 @@ class PmoProblem(object):
         (T, s) = self.B.to_drg_transform
 
         self._sign_constraints = [
-            cp.multiply(self._big_M_r, self.d) + T @ self.m + s <=
-            self._big_M_r - self._epsilon_r,
-            cp.multiply(self._big_M_r, self.d) + T @ self.m + s >=
-            self._epsilon_r,
-            cp.multiply(self._big_M_f, self.d) - v_to_v_con @ self.vs <=
-            self._big_M_f - self._epsilon_f,
-            cp.multiply(self._big_M_f, self.d) - v_to_v_con @ self.vs >=
-            self._epsilon_f
+            cp.multiply(self._big_M_r, self.d) + T @ self.m + s
+            <= self._big_M_r - self._epsilon_r,
+            cp.multiply(self._big_M_r, self.d) + T @ self.m + s >= self._epsilon_r,
+            cp.multiply(self._big_M_f, self.d) - v_to_v_con @ self.vs
+            <= self._big_M_f - self._epsilon_f,
+            cp.multiply(self._big_M_f, self.d) - v_to_v_con @ self.vs
+            >= self._epsilon_f,
         ]
 
         if self._fixed_directions:
             directions = np.zeros((len(self.T.reaction_ids), 1))
-            directions[self.F.lb[self._rxn_idxs_F_to_T, ] >= 0] = 1
+            directions[
+                self.F.lb[
+                    self._rxn_idxs_F_to_T,
+                ]
+                >= 0
+            ] = 1
             self._sign_constraints += [self.d == directions]
 
         # Assemble in a single set of constraints.
@@ -419,10 +532,9 @@ class PmoProblem(object):
             self._flux_lb_constraint,
             self._flux_ub_constraint,
             self._steady_state_constraint,
-            self._m_constraint
+            self._m_constraint,
         ] + self._sign_constraints
 
     def _rebuild_cvxpy_problem(self):
-        """Rebuild the CVXPY representation of the PMO problem.
-        """
+        """Rebuild the CVXPY representation of the PMO problem."""
         self._problem = cp.Problem(self.objective(self), self._constraints)
