@@ -23,12 +23,13 @@ from .gibbs_estimators import EquilibratorGibbsEstimator, GibbsEstimatorInterfac
 from .metabolite import Metabolite
 from .metabolite_interpreter import MetaboliteInterpreter
 from .utils import (
+    apply_transform,
+    covariance_square_root,
     get_candidate_thermodynamic_constraints,
     qrvector,
     qvector,
     to_reactions_ids,
     to_reactions_idxs,
-    transform,
 )
 
 logger = logging.getLogger(__name__)
@@ -53,8 +54,8 @@ class ThermodynamicSpace(object):
         Prior distributions for the metabolite concentrations.
     estimator : GibbsEstimatorInterface, optional
         Object used to estimate Gibbs free energies.
-    drg0_estimate : Optional[Tuple[Q, Q]], optional
-        Estimate of reaction energies (mean and a square root of the covariance matrix)
+    dfg0_estimate : Optional[Tuple[Q, Q]], optional
+        Estimate of formation energies (mean and a square root of the covariance matrix)
         in case the user wants to specify them manually. This is only used if
         :code:`estimator` is :code:`None`.
     """
@@ -64,30 +65,31 @@ class ThermodynamicSpace(object):
         S_constraints: np.array,
         reaction_ids: List[str],
         metabolites: List[Metabolite],
+        metabolite_ids: List[str] = None,
         parameters: CompartmentParameters = None,
         concentrations: ConcentrationsPrior = None,
         estimator: GibbsEstimatorInterface = None,
-        drg0_estimate: Optional[Tuple[Q, Q]] = None
+        dfg0_estimate: Optional[Tuple[Q, Q]] = None,
     ):
         self._S_constraints = S_constraints
         self._reaction_ids = reaction_ids
         self._metabolites = metabolites
-        self._parameters = parameters or CompartmentParameters.load('default')
+        self._parameters = parameters or CompartmentParameters.load("default")
+        self._metabolite_ids = metabolite_ids or [m.key for m in metabolites]
 
-        if estimator is None and drg0_estimate is not None:
-            self._drg0_prime_mean, self._drg0_prime_cov_sqrt = drg0_estimate
+        if estimator is None and dfg0_estimate is not None:
+            self._dfg0_prime_mean, self._dfg0_prime_cov_sqrt = dfg0_estimate
         else:
-            if drg0_estimate is not None:
-                logger.warning('drg0_estimate ignored, since an estimator has '
-                               'has been passed to ThermodynamicSpace.')
+            if dfg0_estimate is not None:
+                logger.warning(
+                    "dfg0_estimate ignored, since an estimator has "
+                    "has been passed to ThermodynamicSpace."
+                )
 
             estimator = estimator or EquilibratorGibbsEstimator()
-            self._drg0_prime_mean, self._drg0_prime_cov_sqrt = \
-                estimator.get_drg0_prime(
-                    self.S_constraints,
-                    metabolites,
-                    self._parameters
-                )
+            self._dfg0_prime_mean, self._dfg0_prime_cov_sqrt = estimator.get_dfg0_prime(
+                self.S_constraints, metabolites, self._parameters
+            )
 
         if not concentrations:
             if isinstance(estimator, EquilibratorGibbsEstimator):
@@ -95,25 +97,44 @@ class ThermodynamicSpace(object):
             else:
                 concentrations = ConcentrationsPrior()
 
-        self._log_conc_mean, self._log_conc_cov = \
-            self._make_concentrations_distribution(metabolites, concentrations)
+        (
+            self._log_conc_mean,
+            self._log_conc_cov,
+        ) = self._make_concentrations_distribution(metabolites, concentrations)
+
+    @property
+    def dfg0_prime_mean(self) -> Q:
+        """Gets the mean of the corrected standard formation energies."""
+        return self._dfg0_prime_mean
+
+    @property
+    def dfg0_prime_cov(self) -> Q:
+        """Gets the covariance of the corrected standard formation energies."""
+        return self._dfg0_prime_cov_sqrt @ self._dfg0_prime_cov_sqrt.T
+
+    @property
+    def dfg0_prime_cov_sqrt(self) -> Q:
+        """Gets a square root of the covariance of the corrected standard
+        formation energies.
+        """
+        return self._dfg0_prime_cov_sqrt
 
     @property
     def drg0_prime_mean(self) -> Q:
         """Gets the mean of the corrected standard reaction energies."""
-        return self._drg0_prime_mean
+        return self.S_constraints.T @ self.dfg0_prime_mean
 
     @property
     def drg0_prime_cov(self) -> Q:
         """Gets the covariance of the corrected standard reaction energies."""
-        return self._drg0_prime_cov_sqrt @ self._drg0_prime_cov_sqrt.T
+        return self.drg0_prime_cov_sqrt @ self.drg0_prime_cov_sqrt.T
 
     @property
     def drg0_prime_cov_sqrt(self) -> Q:
-        """Gets a square root of the covariance of the corrected standard 
+        """Gets a square root of the covariance of the corrected standard
         reaction energies.
         """
-        return self._drg0_prime_cov_sqrt
+        return self.S_constraints.T @ self.dfg0_prime_cov_sqrt
 
     @property
     def log_conc_mean(self):
@@ -125,6 +146,26 @@ class ThermodynamicSpace(object):
         """Gets the covariance of the log-concentrations."""
         return self._log_conc_cov
 
+    @property
+    def dfg_prime_mean(self):
+        """Gets the mean of the formation energies."""
+        return self.dfg0_prime_mean + self.log_conc_mean * R * self.parameters.T()
+
+    @property
+    def dfg_prime_cov(self):
+        """Gets the covariance of the formation energies."""
+        return self.dfg0_prime_cov + self.log_conc_cov * (R * self.parameters.T()) ** 2
+
+    @property
+    def drg_prime_mean(self):
+        """Gets the mean of the reaction energies."""
+        return self.S_constraints.T @ self.dfg_prime_mean
+
+    @property
+    def drg_prime_cov(self):
+        """Gets the covariance of the reaction energies."""
+        return self.S_constraints.T @ self.dfg_prime_cov @ self.S_constraints
+
     @property
     def parameters(self) -> CompartmentParameters:
         """Gets the compartment parameters of the system."""
@@ -137,10 +178,15 @@ class ThermodynamicSpace(object):
 
     @property
     def S_constraints(self) -> np.array:
-        """Gets stoichiometric matrix of the reactions with thermodynamic 
+        """Gets stoichiometric matrix of the reactions with thermodynamic
         constraints."""
         return self._S_constraints
 
+    @property
+    def metabolite_ids(self):
+        """Gets the IDs of the metabolites in the thermodynamic space."""
+        return self._metabolite_ids
+
     @property
     def reaction_ids(self):
         """Gets the IDs of the reactions with thermodynamic constraints."""
@@ -153,7 +199,7 @@ class ThermodynamicSpace(object):
         constrained_rxns: Union[List[int], List[str], cobra.DictList] = None,
         estimator: GibbsEstimatorInterface = None,
         parameters: CompartmentParameters = None,
-        concentrations: ConcentrationsPrior = None
+        concentrations: ConcentrationsPrior = None,
     ) -> ThermodynamicSpace:
         """Constructs a thermodynamic space from a cobrapy model.
 
@@ -182,8 +228,7 @@ class ThermodynamicSpace(object):
             The thermodynamic space for the specified model.
         """
         estimator = estimator or EquilibratorGibbsEstimator()
-        metabolite_interpreter = \
-            metabolite_interpreter or MetaboliteInterpreter()
+        metabolite_interpreter = metabolite_interpreter or MetaboliteInterpreter()
 
         if constrained_rxns is None:
             if isinstance(estimator, EquilibratorGibbsEstimator):
@@ -193,29 +238,35 @@ class ThermodynamicSpace(object):
             constrained_rxns = get_candidate_thermodynamic_constraints(
                 FluxSpace.from_cobrapy_model(model),
                 metabolite_interpreter,
-                ccache=ccache)
+                ccache=ccache,
+            )
 
         constrained_rxn_ids = to_reactions_ids(constrained_rxns, model)
         constrained_rxn_idxs = to_reactions_idxs(constrained_rxns, model)
 
-        S_constraints = create_stoichiometric_matrix(
-            model)[:, constrained_rxn_idxs]
+        S_constraints = create_stoichiometric_matrix(model)[:, constrained_rxn_idxs]
+        is_metabolite_needed = np.sum(np.abs(S_constraints), axis=1) > 0
+        S_constraints = S_constraints[is_metabolite_needed, :]
 
         # Parse metabolites from the model.
         metabolites = []
-        for m in model.metabolites:
-            metabolite = metabolite_interpreter.read(m.id)
-            metabolite.nH = m.elements.get('H', 0)
-            metabolite.z = m.charge
-            metabolites.append(metabolite)
+        metabolite_ids = []
+        for i, m in enumerate(model.metabolites):
+            if is_metabolite_needed[i]:
+                metabolite = metabolite_interpreter.read(m.id)
+                metabolite.nH = m.elements.get("H", 0)
+                metabolite.z = m.charge
+                metabolites.append(metabolite)
+                metabolite_ids.append(m.id)
 
         return ThermodynamicSpace(
             S_constraints,
             constrained_rxn_ids,
             metabolites,
+            metabolite_ids,
             parameters,
             concentrations,
-            estimator
+            estimator,
         )
 
     def _make_concentrations_distribution(
@@ -223,15 +274,23 @@ class ThermodynamicSpace(object):
         metabolites: List[Metabolite],
         concentrations: ConcentrationsPrior,
     ) -> Tuple[Q, Q]:
-        """Construct the distribution over metabolite concentrations.
-        """
+        """Construct the distribution over metabolite concentrations."""
         # TODO: Support LogUniform
-        log_conc_mean = qvector([
-            Q(concentrations.get(m.key, m.compartment).log_mean)
-            for m in metabolites])
-        log_conc_cov = Q(np.diag(np.square(qrvector([
-            Q(concentrations.get(m.key, m.compartment).log_std)
-            for m in metabolites]).m)))
+        log_conc_mean = qvector(
+            [Q(concentrations.get(m.key, m.compartment).log_mean) for m in metabolites]
+        )
+        log_conc_cov = Q(
+            np.diag(
+                np.square(
+                    qrvector(
+                        [
+                            Q(concentrations.get(m.key, m.compartment).log_std)
+                            for m in metabolites
+                        ]
+                    ).m
+                )
+            )
+        )
 
         return log_conc_mean, log_conc_cov
 
@@ -247,13 +306,13 @@ class ThermodynamicSpaceBasis(object):
         The target thermodynamic space.
     explicit_log_conc : bool, optional
         True if log-concentrations should be represented explicitely in the basis, false
-        otherwise. By default True. 
+        otherwise. By default True.
     explicit_drg0 : bool, optional
         True if standard reaction energies should be represented explicitely in the
         basis, false otherwise. By default True.
     explicit_drg : bool, optional
         True if reaction energies should be represented explicitely in the basis, false
-        otherwise. By default True. 
+        otherwise. By default True.
     min_eigenvalue : float, optional
         Minimum eigenvalue of for a vector to be part of the basis.
     """
@@ -264,18 +323,24 @@ class ThermodynamicSpaceBasis(object):
         explicit_log_conc: bool = True,
         explicit_drg0: bool = True,
         explicit_drg: bool = True,
-        min_eigenvalue: float = default_min_eigenvalue_tds_basis
+        min_eigenvalue: float = default_min_eigenvalue_tds_basis,
     ):
         self._to_log_conc_transform: Optional[Tuple[np.array, np.array]] = None
         self._to_drg0_transform: Optional[Tuple[np.array, np.array]] = None
         self._to_drg_transform: Optional[Tuple[np.array, np.array]] = None
 
+        self._observables_ranges: Dict[str, List[int]] = {
+            "drg": [],
+            "drg0": [],
+            "log_conc": [],
+        }
+
         self._compute_basis(
             thermodynamic_space,
             explicit_log_conc,
             explicit_drg0,
             explicit_drg,
-            min_eigenvalue
+            min_eigenvalue,
         )
 
     @property
@@ -296,9 +361,19 @@ class ThermodynamicSpaceBasis(object):
         if reaction energies are not represented explicitely."""
         return self._to_drg_transform
 
+    @property
+    def to_observables_transform(self) -> Tuple[np.array, np.array]:
+        """Gets the transformation from this space to the selected observables."""
+        return self._to_observables_transform
+
+    @property
+    def observables_ranges(self) -> Dict[str, List[int]]:
+        """The ranges of the different variables in a vector of observables."""
+        return self._observables_ranges
+
     @property
     def sigmas(self) -> np.array:
-        """Gets a vector containing the standard deviation of each variable in 
+        """Gets a vector containing the standard deviation of each variable in
         the minimal basis."""
         return self._sigmas
 
@@ -309,7 +384,7 @@ class ThermodynamicSpaceBasis(object):
 
     def to_log_conc(self, vars: np.array) -> np.array:
         """Transform a vector or matrix in the basis to a vector or matrix of
-        log-concentrations. 
+        log-concentrations.
 
         Parameters
         ----------
@@ -320,16 +395,16 @@ class ThermodynamicSpaceBasis(object):
         -------
         np.array
             The transformed vector or matrix. :code:`None` if concentrations are not
-            represented explicitely. 
+            represented explicitely.
         """
         if not self.to_log_conc_transform:
             return None
         else:
-            return transform(vars, self.to_log_conc_transform)
+            return apply_transform(vars, self.to_log_conc_transform)
 
     def to_drg0(self, vars: np.array) -> np.array:
         """Transform a vector or matrix in the basis to a vector or matrix of
-        standard reaction energies. 
+        standard reaction energies.
 
         Parameters
         ----------
@@ -340,16 +415,16 @@ class ThermodynamicSpaceBasis(object):
         -------
         np.array
             The transformed vector or matrix. :code:`None` if standard reaction energies
-            are not represented explicitely. 
+            are not represented explicitely.
         """
         if not self.to_drg0_transform:
             return None
         else:
-            return transform(vars, self.to_drg0_transform)
+            return apply_transform(vars, self.to_drg0_transform)
 
     def to_drg(self, vars: np.array) -> np.array:
         """Transform a vector or matrix in the basis to a vector or matrix of
-        reaction energies. 
+        reaction energies.
 
         Parameters
         ----------
@@ -365,7 +440,7 @@ class ThermodynamicSpaceBasis(object):
         if not self.to_drg_transform:
             return None
         else:
-            return transform(vars, self.to_drg_transform)
+            return apply_transform(vars, self.to_drg_transform)
 
     def _compute_basis(
         self,
@@ -373,48 +448,41 @@ class ThermodynamicSpaceBasis(object):
         explicit_log_conc,
         explicit_drg0,
         explicit_drg,
-        tolerance
+        tolerance,
     ):
         """Compute the basis of the thermodynamic space, only making explicit the
         specified variables.
         """
         log_conc_mean = thermodynamic_space.log_conc_mean.m
         log_conc_cov = thermodynamic_space.log_conc_cov.m
-        drg0_prime_mean = thermodynamic_space.drg0_prime_mean.m_as('kJ/mol')
-        drg0_prime_cov_sqrt = thermodynamic_space.drg0_prime_cov_sqrt.m_as(
-            'kJ/mol')
+        drg0_prime_mean = thermodynamic_space.drg0_prime_mean.m_as("kJ/mol")
+        drg0_prime_cov_sqrt = thermodynamic_space.drg0_prime_cov_sqrt.m_as("kJ/mol")
         S_con_T = thermodynamic_space.S_constraints.T
-        RT = (R * thermodynamic_space.parameters.T()).m_as('kJ/mol')
+        RT = (R * thermodynamic_space.parameters.T()).m_as("kJ/mol")
         n_mets = log_conc_mean.size
         n_rxns = drg0_prime_mean.size
 
         # Use the eigenvalue decomposition to find the minimal basis for the
         # log concentrations.
-        log_conc_cov_sqrt = self._get_minimal_subspace(
-            log_conc_cov, tolerance)
+        log_conc_cov_sqrt = covariance_square_root(log_conc_cov, tolerance)
 
         # Combined mean and squared root of the covariance of log-concentrations
         # and DrG'°.
-        log_conc_drg0_mean = np.block([
-            [log_conc_mean],
-            [drg0_prime_mean]
-        ])
+        log_conc_drg0_mean = np.block([[log_conc_mean], [drg0_prime_mean]])
 
         log_conc_drg0_cov_sqrt = scipy.linalg.block_diag(
-            log_conc_cov_sqrt, drg0_prime_cov_sqrt)
+            log_conc_cov_sqrt, drg0_prime_cov_sqrt
+        )
 
         # Construct the transform from log-concentrations and DrG'° to the
         # variables that must be explicitly represented ("observables").
         mapping_blocks = []
         if explicit_log_conc:
-            mapping_blocks.append(
-                [np.identity(n_mets),       np.zeros((n_mets, n_rxns))])
+            mapping_blocks.append([np.identity(n_mets), np.zeros((n_mets, n_rxns))])
         if explicit_drg0:
-            mapping_blocks.append(
-                [np.zeros((n_rxns, n_mets)),  np.identity(n_rxns)])
+            mapping_blocks.append([np.zeros((n_rxns, n_mets)), np.identity(n_rxns)])
         if explicit_drg:
-            mapping_blocks.append(
-                [RT * S_con_T,              np.identity(n_rxns)])
+            mapping_blocks.append([RT * S_con_T, np.identity(n_rxns)])
 
         to_observables = np.block(mapping_blocks)
         observables_mean = to_observables @ log_conc_drg0_mean
@@ -422,9 +490,9 @@ class ThermodynamicSpaceBasis(object):
 
         # Find a square root for the covariance matrix with minimum
         # dimensionality using the QR decomposition.
-        observables_cov_sqrt = LINALG.qr_rank_deficient(
-            observables_cov_sqrt.T).T
+        observables_cov_sqrt = LINALG.qr_rank_deficient(observables_cov_sqrt.T).T
         observables_cov_sqrt[np.abs(observables_cov_sqrt) < 1e-5] = 0
+        self._to_observables_transform = (observables_cov_sqrt, observables_mean)
         self._dimensionality = observables_cov_sqrt.shape[1]
         self._sigmas = la.norm(observables_cov_sqrt, axis=0).T
 
@@ -432,42 +500,28 @@ class ThermodynamicSpaceBasis(object):
         # observables.
         current_id = 0
         if explicit_log_conc:
+            self._observables_ranges["log_conc"] = list(
+                range(current_id, current_id + n_mets)
+            )
             self._to_log_conc_transform = (
-                observables_cov_sqrt[current_id:current_id+n_mets, :],
-                observables_mean[current_id:current_id+n_mets]
+                observables_cov_sqrt[self.observables_ranges["log_conc"], :],
+                observables_mean[self.observables_ranges["log_conc"]],
             )
             current_id += n_mets
         if explicit_drg0:
+            self._observables_ranges["drg0"] = list(
+                range(current_id, current_id + n_rxns)
+            )
             self._to_drg0_transform = (
-                observables_cov_sqrt[current_id:current_id+n_rxns, :],
-                observables_mean[current_id:current_id+n_rxns]
+                observables_cov_sqrt[self.observables_ranges["drg0"], :],
+                observables_mean[self.observables_ranges["drg0"]],
             )
             current_id += n_rxns
         if explicit_drg:
+            self._observables_ranges["drg"] = list(
+                range(current_id, current_id + n_rxns)
+            )
             self._to_drg_transform = (
-                observables_cov_sqrt[current_id:current_id+n_rxns, :],
-                observables_mean[current_id:current_id+n_rxns]
+                observables_cov_sqrt[self.observables_ranges["drg"], :],
+                observables_mean[self.observables_ranges["drg"]],
             )
-
-    def _get_minimal_subspace(
-        self,
-        covariance: np.array,
-        min_eigenvalue
-    ) -> np.array:
-        """Gets a full-rank square root of a covariance matrix.
-        """
-        w, v = la.eigh(covariance)
-        assert np.all(np.abs(np.imag(w)) < min_eigenvalue) and \
-            np.all(np.abs(np.imag(v * np.real(w))) < min_eigenvalue), 'The ' \
-            'eigenvalue decomposition of the covariance matrix has ' \
-            'imaginary parts larger than the specified minimum eigenvalue.'
-        assert np.min(w) > -min_eigenvalue, 'The covariance matrix has ' \
-            'negative eigenvalues larger than the specified minimum eigenvalue.'
-
-        w = np.real(w)
-        v = np.real(v)
-        selected_dims = w >= min_eigenvalue
-        w = w[selected_dims]
-        v = v[:, selected_dims]
-
-        return v @ np.diag(np.sqrt(w))
